from importlib import import_module

from django.conf import settings
from django.contrib.sessions.base_session import AbstractBaseSession, BaseSessionManager
from django.core.cache import caches
from django.db import models
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _

import qsessions.geoip as geoip


class SessionQuerySet(models.QuerySet):
    def delete(self):
        """
        Delete sessions from both DB and cache (first cache, then DB)
        """
        SessionStore = Session.get_session_store_class()
        prefix = getattr(SessionStore, 'cache_key_prefix', None)
        if prefix is not None:
            caches[settings.SESSION_CACHE_ALIAS].delete_many(prefix + s.session_key for s in self)
        return super(SessionQuerySet, self).delete()


class SessionManager(BaseSessionManager.from_queryset(SessionQuerySet)):
    use_in_migrations = True


class Session(AbstractBaseSession):
    """
    Session objects containing user session information.
    """
    user = models.ForeignKey(getattr(settings, 'AUTH_USER_MODEL', 'auth.User'),
                             null=True, on_delete=models.CASCADE)
    user_agent = models.CharField(null=True, blank=True, max_length=300)
    created_at = models.DateTimeField(null=True)
    updated_at = models.DateTimeField(auto_now=True)
    ip = models.GenericIPAddressField(null=True, blank=True, verbose_name=_('IP'))

    objects = SessionManager()

    @classmethod
    def get_session_store_class(cls):
        return import_module(settings.SESSION_ENGINE).SessionStore

    def save(self, *args, **kwargs):
        # FIXME: find a better solution for `created_at` field which does not need an extra query.
        # https://code.djangoproject.com/ticket/17654
        try:
            self.created_at = Session.objects.get(pk=self.pk).created_at
        except Session.DoesNotExist:
            self.created_at = timezone.now()
        super(Session, self).save(*args, **kwargs)

    def delete(self, *args, **kwargs):
        """
        Delete session from both DB and cache (first cache, then DB)
        """
        SessionStore = Session.get_session_store_class()
        prefix = getattr(SessionStore, 'cache_key_prefix', None)
        if prefix is not None:
            caches[settings.SESSION_CACHE_ALIAS].delete(prefix + self.session_key)
        return super(Session, self).delete(*args, **kwargs)

    def location(self):
        return geoip.ip_to_location(self.ip)

    def location_info(self):
        return geoip.ip_to_location_info(self.ip)

    def device(self):
        """
        Describe the user agent of this session, if any
        :rtype: user_agents.parsers.UserAgent | None
        """
        if self.user_agent:
            import user_agents  # late import to avoid import cost
            return user_agents.parse(self.user_agent)
        return None



from django.utils.translation import ugettext_lazy as _
from django.contrib.auth import get_user_model
from django.utils.html import format_html
from django.utils.timezone import now
from django.contrib import admin
from .models import Session
from pprint import pformat


class ExpiredFilter(admin.SimpleListFilter):
    title = _('Is Valid')
    parameter_name = 'active'

    def lookups(self, request, model_admin):
        return (
            ('1', _('Active')),
            ('0', _('Expired'))
        )

    def queryset(self, request, queryset):
        if self.value() == '1':
            return queryset.filter(expire_date__gt=now())
        elif self.value() == '0':
            return queryset.filter(expire_date__lte=now())


class OwnerFilter(admin.SimpleListFilter):
    title = _('Owner')
    parameter_name = 'owner'

    def lookups(self, request, model_admin):
        return (
            ('my', _('Self')),
        )

    def queryset(self, request, queryset):
        if self.value() == 'my':
            return queryset.filter(user=request.user)


@admin.register(Session)
class SessionAdmin(admin.ModelAdmin):
    list_display = ('ip', 'user', 'is_valid', 'created_at', 'expire_date', 'device', 'location')
    readonly_fields = ('ip', 'location', 'user', 'user_agent', 'is_valid', 'expire_date',
                       'created_at', 'updated_at', 'user_agent', 'device', 'session_key',
                       'session_data_decoded')
    list_filter = ExpiredFilter, OwnerFilter
    fields = ('user', 'ip', 'location', 'is_valid', 'created_at', 'updated_at', 'expire_date',
              'user_agent', 'device', 'session_key', 'session_data_decoded')
    ordering = ('-expire_date', )

    def get_search_fields(self, request):
        User = get_user_model()
        return 'ip', 'user__%s' % getattr(User, 'USERNAME_FIELD', 'username'),\
               'user__%s' % getattr(User, 'USERNAME_EMAIL', 'email')

    def is_valid(self, obj):
        return obj.expire_date > now()

    is_valid.boolean = True
    is_valid.short_description = _('Is valid')

    def session_data_decoded(self, obj):
        return format_html(
            '<pre style="white-space: pre-wrap; max-width: 800px; display: inline-block; direction: ltr;">{}</pre>',
            pformat(obj.get_decoded()))

    session_data_decoded.short_description = _('Session data')



from django.contrib.sessions.backends.cached_db import SessionStore as CachedDBStore
from .common import QSessionStoreMixin

KEY_PREFIX = "qsessions.q_cached_db"


class SessionStore(QSessionStoreMixin, CachedDBStore):
    """
    Implements cached, database backed sessions, with a foreign key to User.
    It also stores IP and User Agent.
    """
    cache_key_prefix = KEY_PREFIX



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2018-02-05 06:55
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('qsessions', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='session',
            name='created_at',
            field=models.DateTimeField(null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-12-19 16:00
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import qsessions.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Session',
            fields=[
                ('session_key', models.CharField(max_length=40, primary_key=True, serialize=False, verbose_name='session key')),
                ('session_data', models.TextField(verbose_name='session data')),
                ('expire_date', models.DateTimeField(db_index=True, verbose_name='expire date')),
                ('user_agent', models.CharField(blank=True, max_length=300, null=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('ip', models.GenericIPAddressField(blank=True, null=True, verbose_name='IP')),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'session',
                'abstract': False,
                'verbose_name_plural': 'sessions',
            },
            managers=[
                ('objects', qsessions.models.SessionManager()),
            ],
        ),
    ]



from __future__ import unicode_literals

import warnings
from django.contrib.gis.geoip2 import HAS_GEOIP2


def ip_to_location_info(ip):
    """
    Get a dictionary of location info for a given IP address.

    The format of the dictionary is the same provided by the functions
    in django.contrib.gis.geoip2.base.GeoIP2.
    """

    if not HAS_GEOIP2:
        return None

    from django.contrib.gis.geoip2 import GeoIP2
    try:
        g = GeoIP2()
    except Exception as e:
        warnings.warn(str(e))
        return None

    try:
        return g.city(ip)
    except Exception:
        try:
            return g.country(ip)
        except Exception as e:
            warnings.warn(str(e))


def ip_to_location(ip):
    """
    Transform an IP address into an approximate location.

    Example output:

    * Zwolle, The Netherlands
    * The Netherlands
    * None
    """
    loc = ip_to_location_info(ip)
    if not loc:
        return None

    if loc.get('country_name'):
        if loc.get('city'):
            return '{}, {}'.format(loc['city'], loc['country_name'])
        return loc['country_name']

    return None



from django.core.management.base import BaseCommand
from django.conf import settings
import urllib.request
import tarfile
import os


class Command(BaseCommand):
    help = 'Update GeoIP2 database'

    def handle(self, verbosity=0, **options):
        db_path = getattr(settings, 'GEOIP_PATH', None)
        if not db_path:
            if verbosity >= 1:
                self.stderr.write("No GEOIP_PATH defined, not downloading database.")
            return

        if not os.path.exists(db_path):
            os.makedirs(db_path)

        for url in [
            "http://geolite.maxmind.com/download/geoip/database/GeoLite2-City.tar.gz",
            "http://geolite.maxmind.com/download/geoip/database/GeoLite2-Country.tar.gz",
        ]:
            filename = os.path.join(db_path, os.path.basename(url))
            if verbosity >= 1:
                self.stdout.write('Downloading and extracting {url}...'.format(url=url))
            urllib.request.urlretrieve(url, filename)
            self.extract_tar(db_path, filename)
            os.remove(filename)

    @staticmethod
    def extract_tar(db_path, tar_path):
        with tarfile.open(tar_path) as tarball:
            for tarinfo in tarball:
                if tarinfo.name.endswith('.mmdb'):
                    tarinfo.name = os.path.basename(tarinfo.name)
                    tarball.extract(tarinfo, path=db_path)



from django.contrib.sessions.backends.db import SessionStore as DBStore
from .common import QSessionStoreMixin


class SessionStore(QSessionStoreMixin, DBStore):
    """
    Implements database backed sessions, with a foreign key to User.
    It also stores IP and User Agent.
    """
    pass



from django.contrib.sessions.middleware import SessionMiddleware as DjSessionMiddleware
from django.conf import settings
from ipware.ip import get_real_ip, get_ip


class SessionMiddleware(DjSessionMiddleware):

    def process_request(self, request):
        session_key = request.COOKIES.get(settings.SESSION_COOKIE_NAME)
        request.session = self.SessionStore(
            ip=get_real_ip(request) or get_ip(request),
            user_agent=request.META.get('HTTP_USER_AGENT', ''),
            session_key=session_key
        )



import os
from setuptools import setup, find_packages

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

dev_requirements = [
    'geoip2==3.0.0',  # for testing GeoIP2
    'pytest',
    'pytest-cov',
    'pytest-django',
]

setup(
    name='django-qsessions',
    version='0.4.0',
    description='Extended session backends for Django',
    long_description=README,
    author='Mohammad Javad Naderi',
    url='https://github.com/QueraTeam/django-qsessions',
    download_url='https://pypi.python.org/pypi/django-qsessions',
    license='MIT',
    packages=find_packages('.', include=('qsessions', 'qsessions.*')),
    include_package_data=True,
    install_requires=[
        'Django>=1.10',
        'user-agents>=1.1.0',
        'django-ipware>=1.1.5',
    ],
    extras_require={'dev': dev_requirements},
    tests_require=dev_requirements,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP :: Session',
        'Topic :: Security',
    ],
)



from qsessions import IP_SESSION_KEY, USER_AGENT_SESSION_KEY
from django.contrib import auth


class QSessionStoreMixin(object):
    """
    Implements sessions stored with:
    - Foreign key to User
    - IP address
    - User Agent
    """
    def __init__(self, session_key=None, user_agent=None, ip=None):
        self.user_agent = user_agent[:300] if user_agent else user_agent
        self.ip = ip
        super(QSessionStoreMixin, self).__init__(session_key)

    @classmethod
    def get_model_class(cls):
        from qsessions.models import Session
        return Session

    def load(self):
        data = super(QSessionStoreMixin, self).load()
        if data.get(USER_AGENT_SESSION_KEY) != self.user_agent \
                or data.get(IP_SESSION_KEY) != self.ip:
            # If IP or User Agent has changed, set modified to True in order to save
            # the new IP and User Agent
            self.modified = True
        return data

    def save(self, must_create=False):
        # Store IP and User Agent in session_data
        if USER_AGENT_SESSION_KEY not in self or self[USER_AGENT_SESSION_KEY] != self.user_agent:
            self[USER_AGENT_SESSION_KEY] = self.user_agent
        if IP_SESSION_KEY not in self or self[IP_SESSION_KEY] != self.ip:
            self[IP_SESSION_KEY] = self.ip
        super(QSessionStoreMixin, self).save(must_create)

    def create_model_instance(self, data):
        # Store User, User Agent, and IP in Session (in DB).
        return self.model(
            session_key=self._get_or_create_session_key(),
            session_data=self.encode(data),
            expire_date=self.get_expiry_date(),
            user_id=self.get(auth.SESSION_KEY),
            user_agent=self.user_agent,
            ip=self.ip
        )
